/* {{{ Copyright etc. */

/*  feat_model - create FEAT design matrix and contrast file

    Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 1999-2002 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    
    LICENCE
    
    FMRIB Software Library, Release 3.0 (c) 2002, The University of Oxford
    (the "Software")
    
    The Software remains the property of the University of Oxford ("the
    University").
    
    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.
    
    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.
    
    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.
    
    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Isis
    Innovation Limited ("Isis"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    innovation@isis.ox.ac.uk quoting reference DE/1112. */

/* }}} */
/* {{{ includes and defines */

#include "libss/libss.h"

#define DT           0.05 /* temporal sampling for initial model, in seconds */
#define MINCONDITION 5e-2 /* used to test ratio of min:max eigenvalues after SVD of matrix, for rank testing */
#define NEGSECS      30   /* amount of negative modelling allowed, for custom 3 etc, in seconds */

/* }}} */
/* {{{ gammpdf */

float gammpdf(float h, float l, float x)
{
        int i;
        double W=h+5.5,X=h,Y=h,Z=1.000000000190015,gammaln;
        static double cof[6]={76.18009172947146,-86.50532032941677,
                24.01409824083091,-1.231739572450155,
                0.1208650973866179e-2,-0.5395239384953e-5};

	if (x<=0.0 || h<=0.0 || l<=0.0) return 0.0;

        W -= (X+0.5)*log(W);
        Z=1.000000000190015;
        for (i=0;i<=5;i++) Z += cof[i]/++Y;
        gammaln=-W+log(2.5066282746310005*Z/X);

	/* equation is   pow(l,h) * pow(x,h-1) * exp(-l*x - gammln)  but replace with more numerically sensible equivalent */
        return exp( h*log(l) + (h-1)*log(x) - l*x - gammaln );
}

/* }}} */
/* {{{ freeSVD */

/* This SVD routine is based on pgs 30-48 of "Compact Numerical Methods
   for Computers" by J.C. Nash (1990), used to compute the pseudoinverse.
   Modifications include:
        Translation from Pascal to ANSI C.
        Array indexing from 0 rather than 1.
        Float replaced by double everywhere.
        Support for the Matrix structure.
        I changed the array indexing so that the matricies (float [][])
           could be replaced be a single list (double *) for more
           efficient communication with Mathematica.
*/

#define TOLERANCE 1.0e-22

void SVD(double *W, double *Z, int nRow, int nCol)
{
  int i, j, k, EstColRank, RotCount, SweepCount, slimit;
  double eps, e2, tol, vt, p, x0, y0, q, r, c0, s0, d1, d2;
  eps = TOLERANCE;
  slimit = nCol/4;
  if (slimit < 6.0)
    slimit = 6;
  SweepCount = 0;
  e2 = 10.0*nRow*eps*eps;
  tol = eps*.1;
  EstColRank = nCol;
  for (i=0; i<nCol; i++)
    for (j=0; j<nCol; j++)
      {
        W[nCol*(nRow+i)+j] = 0.0;
        W[nCol*(nRow+i)+i] = 1.0;
      }
  RotCount = EstColRank*(EstColRank-1)/2;
  while (RotCount != 0 && SweepCount <= slimit)
    {
      RotCount = EstColRank*(EstColRank-1)/2;
      SweepCount++;
      for (j=0; j<EstColRank-1; j++)
        {
          for (k=j+1; k<EstColRank; k++)
            {
              p = q = r = 0.0;
              for (i=0; i<nRow; i++)
                {
                  x0 = W[nCol*i+j]; y0 = W[nCol*i+k];
                  p += x0*y0; q += x0*x0; r += y0*y0;
                }
              Z[j] = q; Z[k] = r;
              if (q >= r)
                {
                  if (q<=e2*Z[0] || fabs(p)<=tol*q) RotCount--;
                  else
                    {
                      p /= q; r = 1 - r/q; vt = sqrt(4*p*p+r*r);
                      c0 = sqrt(fabs(.5*(1+r/vt))); s0 = p/(vt*c0);
                      for (i=0; i<nRow+nCol; i++)
                        {
                          d1 = W[nCol*i+j]; d2 = W[nCol*i+k];
                          W[nCol*i+j] = d1*c0+d2*s0; W[nCol*i+k] = -d1*s0+d2*c0;
                        }
                    }
                }
              else
                {
                  p /= r; q = q/r-1; vt = sqrt(4*p*p+q*q);
                  s0 = sqrt(fabs(.5*(1-q/vt)));
                  if (p<0) s0 = -s0;
                  c0 = p/(vt*s0);
                  for (i=0; i<nRow+nCol; i++)
                    {
                      d1 = W[nCol*i+j]; d2 = W[nCol*i+k];
                      W[nCol*i+j] = d1*c0+d2*s0; W[nCol*i+k] = -d1*s0+d2*c0;
                    }
                }
            }
        }
      while (EstColRank>=3 &&
	     Z[EstColRank-1]<=Z[0]*tol+tol*tol)
        EstColRank--;
    }
}

/* }}} */
/* {{{ renorm kernel */

void renorm_kernel(float *cX, int fw)
{
  int i;
  double kernelsum=0;

  for(i=0; i<fw; i++)
    kernelsum += cX[i]; /* maybe this should be fabs(cX[i]), but looking at double-gamma output, I've not done this */

  for(i=0; i<fw; i++)
    cX[i] /= kernelsum;
}

/* }}} */
/* {{{ carry out the convolution */

float *do_convolve(float *input, float *kernel, int mnpts, int fw, int phase)
{
  int i, t;
  float *oX=calloc(mnpts,sizeof(float));

  for(t=0;t<mnpts;t++)
    for(i=MAX(0,t+phase-mnpts); i<MIN(fw,t+phase+1); i++)
      oX[t] += input[t+phase-i] * kernel[i];

  return oX;
}

/* }}} */
/* {{{ resample down in time */

/* sample in the MIDDLE of the upsampled period */

void do_resample(float *input, float *output, int npts, float trmult, int negpts)
{
  int t;

  for(t=0;t<npts;t++)
    output[t] = input[((int)((t+0.5)*trmult))+negpts];
}

/* }}} */
/* {{{ find_line */

/* finds LAST matching entry in setup file */

char *find_line(char *filename, char *key, char *fl)
{
  FILE *fd;
  char *return_ptr=NULL, tmp_fl[10000];
  int  j;

  fd=fopen(filename,"rb");

  while ( fgets(tmp_fl, 1000, fd) != NULL )
    if (strncmp(tmp_fl,"set ",4)==0)
      for(j=4; tmp_fl[j]!=0; j++)
	if (strncmp(key,tmp_fl+j,strlen(key))==0)
	  {
	    strcpy(fl,tmp_fl+j);
	    return_ptr = fl+1+strlen(key);
	    if (fl[strlen(fl)-1]==10) fl[strlen(fl)-1]=0;
	    if (fl[strlen(fl)-1]=='"') fl[strlen(fl)-1]=0;
	    if (fl[1+strlen(key)]=='"') return_ptr++;
	  }

  fclose(fd);

  if (return_ptr==NULL)
    {
      printf("Can't find key %s\n",key);
      exit(1);
    }
  else
    return return_ptr;
}

/* }}} */
/* {{{ setup_font */

/* taken (I think) from pbmtext etc by Jef Poskanzer and George Phillips */

#define DEFAULTFONT_ROWS 155
#define DEFAULTFONT_COLS 112

#define FONT_WIDTH       7
#define FONT_HEIGHT      13

#define FONT_Y_SIZE     30

typedef struct { unsigned char font[DEFAULTFONT_ROWS][DEFAULTFONT_COLS];
                 int           char_row0[95], char_col0[95],
                               char_width, char_height; } FONT_DATA;

void error_exit(char *outkey )
{
  printf(outkey);
  exit(1);
}

void setup_font(FONT_DATA *font_data)
{
    /* {{{ Default Font */

/* The default font, packed in hex so this source file doesn't get huge.
   You can replace this with your own font using pbm_dumpfont().
*/

static unsigned long defaultfont_bits[DEFAULTFONT_ROWS][(DEFAULTFONT_COLS+31)/32] = {
    {0x00000000,0x20000c00,0x10000000,0x00000000},
    {0xc600a000,0x42000810,0x00000002,0x00000063},
    {0x6c00a000,0x45000810,0x00000002,0x00000036},
    {0x6c00a000,0x88800808,0xf2e1dee2,0x00000036},
    {0x54000000,0x80000800,0x11122442,0x0000002a},
    {0x54000001,0x00000800,0x11122442,0x0000002a},
    {0x54000001,0x00000800,0x11122282,0x0000002a},
    {0x44000102,0x00000800,0x11122382,0x00000022},
    {0xee000102,0x00000800,0x11e1e102,0x00000077},
    {0x00000204,0x00000800,0x11002102,0x00000000},
    {0x00000000,0x00000c00,0x11002102,0x00000000},
    {0x00000000,0x003f8000,0xe3807600,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x02000080,0x00040000,0x00120000,0x00000001},
    {0x04000082,0x828e1838,0x20210100,0x00000002},
    {0x04000082,0x82912448,0x20210100,0x00000002},
    {0x08000082,0x8fd01940,0x404087c2,0x00000004},
    {0x08000080,0x050c0622,0x00408102,0x00000004},
    {0x10000080,0x05061874,0x0040828f,0x00008008},
    {0x10000080,0x1f912688,0x00408002,0x00000008},
    {0x20000000,0x0a11098c,0x00408002,0x00000010},
    {0x20000080,0x0a0e0672,0x00210000,0x00000010},
    {0x40000000,0x00040000,0x00210000,0x00000020},
    {0x00000000,0x00000000,0x00120000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x004e0838,0x7023e1cf,0x00008000},
    {0x00000000,0x00913844,0x88620208,0x00008000},
    {0x08000000,0x00910844,0x08a20401,0x00000004},
    {0x10000000,0x01110844,0x08a20401,0x00000008},
    {0x20000000,0x01110808,0x3123c781,0x00000010},
    {0x400003e0,0x02110810,0x0a202441,0x00000020},
    {0x20000000,0x02110820,0x0bf02442,0x00000010},
    {0x10008000,0x04110844,0x88242442,0x00000008},
    {0x08008002,0x040e3e7c,0x7073c382,0x00000004},
    {0x00010000,0x08000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x0000e1c0,0x00000000,0x00000000,0x00000000},
    {0x00011220,0x00000000,0x70e38f87,0x00000000},
    {0x20011220,0x00020020,0x89108448,0x00008010},
    {0x10011220,0x00040010,0x09314448,0x00008008},
    {0x0800e221,0x02083e08,0x11514788,0x00000004},
    {0x040111e0,0x00100004,0x2153e448,0x00000002},
    {0x08011020,0x00083e08,0x213a2448,0x00008004},
    {0x10011040,0x02040010,0x01022448,0x00008008},
    {0x2000e381,0x02020020,0x20e77f87,0x00000010},
    {0x00000000,0x04000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x3803e7ef,0xc73bbe3d,0xdb863ce7,0x0000001c},
    {0x44011224,0x48910808,0x91036648,0x00008022},
    {0x4c011285,0x48910808,0xa1036648,0x00008026},
    {0x54011387,0x081f0808,0xc102a548,0x0000802a},
    {0x54011285,0x09910808,0xe102a548,0x0000802a},
    {0x4e011204,0x08910848,0x9112a4c8,0x00008027},
    {0x40011224,0x08910848,0x891224c8,0x00008020},
    {0x3803e7ef,0x073bbe31,0xcff77e47,0x0000001c},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000003,0x00000000},
    {0x0003e1cf,0x87bff7ef,0xdfbf77c2,0x00000000},
    {0x00013224,0x48a4a244,0x89122442,0x00000000},
    {0x00011224,0x4824a244,0xa8a14482,0x00000000},
    {0x00013227,0x8e04226c,0xa8414102,0x00000000},
    {0x0001e224,0x83842228,0xa8a08102,0x00000000},
    {0x00010224,0x40842228,0xd8a08242,0x00000000},
    {0x00010224,0x48843638,0x51108442,0x00000000},
    {0x0003c1ce,0x6f1f1c10,0x53b9c7c2,0x00000000},
    {0x00000060,0x00000000,0x00000002,0x00000000},
    {0x00000000,0x00000000,0x00000003,0x00000000},
    {0xfe000000,0x00000000,0x00000000,0x0000007f},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00010180,0x000000c0,0x003001c0,0x00000000},
    {0x08008081,0x00040040,0x00100200,0x00000004},
    {0x10008082,0x80040040,0x00100200,0x00000008},
    {0x10004084,0x40023c78,0x70f1c7c7,0x00004008},
    {0x10004080,0x00000244,0x89122208,0x00008008},
    {0x20002080,0x00001e44,0x8113e208,0x00008010},
    {0x10002080,0x00002244,0x81120208,0x00008008},
    {0x10001080,0x00002244,0x89122208,0x00008008},
    {0x10001080,0x00001db8,0x70e9c787,0x00008008},
    {0x10000880,0x00000000,0x00000000,0x00008008},
    {0x08000180,0x00000000,0x00000000,0x00008004},
    {0x00000000,0x1fc00000,0x00000007,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00030080,0x981c0000,0x00000000,0x00000000},
    {0x20010000,0x08040000,0x00000000,0x00000010},
    {0x10010000,0x08040000,0x00000000,0x00000008},
    {0x10016387,0x898474b8,0x72e1d5c7,0x00000008},
    {0x10019080,0x8a042a64,0x89122208,0x00008008},
    {0x08011080,0x8c042a44,0x89122207,0x00000004},
    {0x10011080,0x8a042a44,0x89122200,0x00008008},
    {0x10011080,0x89042a44,0x89122208,0x00008008},
    {0x1003bbe0,0x98dfebe6,0x71e1e787,0x00000008},
    {0x10000000,0x80000000,0x01002000,0x00000008},
    {0x20000000,0x80000000,0x01002000,0x00000010},
    {0x00000007,0x00000000,0x03807000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00008000,0x00000000,0x10410000,0x00000000},
    {0x00008000,0x00000000,0x20408000,0x00000000},
    {0x0001f66e,0xfdfbf77c,0x20408000,0x00000000},
    {0x24008224,0x488a2248,0x20408240,0x00000012},
    {0x54008224,0x4a842210,0x40404540,0x0000002a},
    {0x48008222,0x8a8a1420,0x20408480,0x00000024},
    {0x00008a23,0x85111c44,0x20408000,0x00000000},
    {0x000071d1,0x0531887c,0x20408000,0x00000000},
    {0x00000000,0x00000800,0x20408000,0x00000000},
    {0x00000000,0x00000800,0x10410000,0x00000000},
    {0x00000000,0x00003000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x00000000,0x00000000,0x00000000},
    {0x00000000,0x20000c00,0x10000000,0x00000000},
    {0xc600a000,0x42000810,0x00000002,0x00000063},
    {0x6c00a000,0x45000810,0x00000002,0x00000036},
    {0x6c00a000,0x88800808,0xf2e1dee2,0x00000036},
    {0x54000000,0x80000800,0x11122442,0x0000002a},
    {0x54000001,0x00000800,0x11122442,0x0000002a},
    {0x54000001,0x00000800,0x11122282,0x0000002a},
    {0x44000102,0x00000800,0x11122382,0x00000022},
    {0xee000102,0x00000800,0x11e1e102,0x00000077},
    {0x00000204,0x00000800,0x11002102,0x00000000},
    {0x00000000,0x00000c00,0x11002102,0x00000000},
    {0x00000000,0x003f8000,0xe3807600,0x00000000}
    };

/* }}} */
    unsigned char b;
    int rows, cols,
        scol, brow, d,
        bcol, ch;
    unsigned long l;

    /* {{{ text explanation */

    /*
    ** This routine expects a font bitmap representing the following text:
    **
    ** (0,0)
    **    M ",/^_[`jpqy| M
    **
    **    /  !"#$%&'()*+ /
    **    < ,-./01234567 <
    **    > 89:;<=>?@ABC >
    **    @ DEFGHIJKLMNO @
    **    _ PQRSTUVWXYZ[ _
    **    { \]^_`abcdefg {
    **    } hijklmnopqrs }
    **    ~ tuvwxyz{|}~  ~
    **
    **    M ",/^_[`jpqy| M
    **
    ** The bitmap must be cropped exactly to the edges.
    **
    ** The dissection works by finding the first blank row and column; that
    ** gives the height and width of the maximum-sized character, which is
    ** not too useful.  But the distance from there to the opposite side is
    ** an integral multiple of the cell size, and that's what we need.  Then
    ** it's just a matter of filling in all the coordinates.
    */

/* }}} */
    /* {{{ create font bits */

  for ( rows = 0; rows < DEFAULTFONT_ROWS; ++rows )
  {
    for ( cols = 0; cols < DEFAULTFONT_COLS; cols += 32 )
    {
      l = defaultfont_bits[rows][cols / 32];
      if (cols + 32 < DEFAULTFONT_COLS)
        ch = cols + 32;
      else
        ch = DEFAULTFONT_COLS;
      for ( scol = ch - 1; scol >= cols; --scol )
      {
        if ( l & 1 )
          font_data -> font[rows][scol] = 1;
        else
          font_data -> font[rows][scol] = 0;
        l >>= 1;
      }
    }
  }

/* }}} */
    /* {{{ Find first blank row */

    for ( brow = 0; brow < DEFAULTFONT_ROWS / 6; ++brow )
	{
	b = font_data -> font[brow][0];
	for ( cols = 1; cols < DEFAULTFONT_COLS; ++ cols )
	    if ( font_data -> font[brow][cols] != b )
		goto nextrow;
	goto gotblankrow;
    nextrow: ;
	}

    error_exit("Couldn't find blank row in font\n");

/* }}} */
    /* {{{ gotblankrow: */

gotblankrow:
    /* Find first blank col. */
    for ( bcol = 0; bcol < DEFAULTFONT_COLS / 8; ++bcol )
	{
	b = font_data -> font[0][bcol];
	for ( rows = 1; rows < DEFAULTFONT_ROWS; ++ rows )
	    if ( font_data -> font[rows][bcol] != b )
		goto nextcol;
	goto gotblankcol;
    nextcol: ;
	}
    error_exit("Couldn't find blank col in font.\n");

/* }}} */
    /* {{{ gotblankcol: */

gotblankcol:
    /* Now compute character cell size. */
    d = DEFAULTFONT_ROWS - brow;
    font_data -> char_height = d / 11;
    if ( font_data -> char_height * 11 != d )
	error_exit("problem computing character cell height");
    d = DEFAULTFONT_COLS - bcol;
    font_data -> char_width = d / 15;
    if ( font_data -> char_width * 15 != d )
	error_exit("problem computing character cell width");

    /*printf("height=%d width=%d\n",font_data -> char_height,
           font_data -> char_width);*/

    /* Now fill in the 0,0 coords. */
    rows = font_data -> char_height * 2;
    cols = font_data -> char_width * 2;
    for ( ch = 0; ch < 95; ++ch )
	{
	font_data -> char_row0[ch] = rows;
	font_data -> char_col0[ch] = cols;
	cols += font_data -> char_width;
	if ( cols >= font_data -> char_width * 14 )
	    {
	    cols = font_data -> char_width * 2;
	    rows += font_data -> char_height;
	    }
	}

/* }}} */
}

/* }}} */
/* {{{ write_string */

void write_string(unsigned char *in, int x, int y, char *the_string,
		  FONT_DATA *font_data, int colour, int x_size, int y_size)
{
  char string_bit;
  int  i, r, c, X, Y;

  for (i=0; the_string[i] != '\0'; i++)
  {
    string_bit = the_string[i] - ' ';
    for ( r = 0; r < font_data->char_height; ++r )
      for ( c = 0; c < font_data->char_width; ++c )
      {
        X = (i*font_data->char_width)+c+x;
        Y = r+y;
        if ( (X>-1) && (X<x_size) && (Y>-1) && (Y<y_size) &&
             (font_data->font[font_data->char_row0[(int)string_bit] + r][font_data->char_col0[(int)string_bit] + c]==1) )
          in[Y*x_size+X]=colour;
      }
  }
}

void write_string_rgb(unsigned char *r, unsigned char *g, unsigned char *b, 
		      int x, int y, char *the_string,
		      FONT_DATA *font_data, int cr, int cg, int cb,
		      int x_size, int y_size)
{
  write_string(r, x, y, the_string, font_data, cr, x_size, y_size);
  write_string(g, x, y, the_string, font_data, cg, x_size, y_size);
  write_string(b, x, y, the_string, font_data, cb, x_size, y_size);
}

/* }}} */

int main(argc,argv)
  int argc;
  char *argv[];
{
  /* {{{ variables */

FILE   *ofp;
int    i, t, con, npts, ndelete, mnpts, level, *negpts,
  orig_evs, orig_ev, real_evs, real_ev, *orig_ev_nreal,
  shape[10000], ncon, nftests, tempfiltyn,
  temphp_yn, templp_yn, f, *F=NULL, *G=NULL;
float  *orig_X, *real_X, *C, tr, mult, trmult, nltffwhm=0, condition, tmpf, *triggers, maxconvwin=0;
double *W, *Z=NULL, minl, maxl;
char   fn[10000], filename[10000], fl[10000], key[10000], the_string[10000], *FSLDIR;
FONT_DATA *font_data = (FONT_DATA*)malloc(sizeof(FONT_DATA));

/* }}} */
  /* {{{ read arguments and prepare variables */

if (argc<2)
{
  printf("Usage: feat_model <design_name_root>\n");
  exit(1);
}

FSLDIR=getenv("FSLDIR");

sprintf(fn,"%s.fsf",argv[1]);

level     = atoi(find_line(fn, "fmri(level)", fl));
tr        = atof(find_line(fn, "fmri(tr)", fl));
npts      = atoi(find_line(fn, "fmri(npts)", fl));
ndelete   = atoi(find_line(fn, "fmri(ndelete)", fl));
orig_evs  = atoi(find_line(fn, "fmri(evs_orig)", fl));
real_evs  = atoi(find_line(fn, "fmri(evs_real)", fl));
ncon      = atoi(find_line(fn, "fmri(ncon_real)", fl));
nftests   = atoi(find_line(fn, "fmri(nftests_real)", fl));

npts    -= ndelete;                       /* subtract off images which have already been deleted */
mult     = 1/DT;                          /* points in high-res model per second */
trmult   = tr*mult;                       /* points in high-res model per TR */
mnpts    = (npts+10)*trmult+NEGSECS*mult; /* number of points in high-res model - add a few on for safety */

orig_X = (float *)calloc(orig_evs*mnpts, sizeof(float));
real_X = (float *)calloc(real_evs*npts,  sizeof(float));

orig_ev_nreal = (int *)malloc(orig_evs*sizeof(int));

setup_font(font_data);

triggers = malloc(orig_evs*2*npts * sizeof(float));
for(i=0;i<orig_evs*2*npts;i++)
     triggers[i]=-1e7;

negpts=(int*)calloc(orig_evs,sizeof(int));

/* }}} */
  /* {{{ read and create design matrix */

if (level==1)
{
  /* {{{ create basic shape */

for(orig_ev=0; orig_ev<orig_evs; orig_ev++)
{
  orig_ev_nreal[orig_ev]=1;

  sprintf(key,"fmri(shape%d)",orig_ev+1); shape[orig_ev]=atoi(find_line(fn, key, fl));

  switch (shape[orig_ev])
    {
    case 0:
      /* {{{ square wave */

{
  float skip, off, on, phase, stop;

  sprintf(key,"fmri(skip%d)",orig_ev+1);  skip=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(off%d)",orig_ev+1);   off=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(on%d)",orig_ev+1);    on=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(phase%d)",orig_ev+1); phase=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(stop%d)",orig_ev+1);  stop=atof(find_line(fn, key, fl));

  skip*=mult; off*=mult; on*=mult; phase*=mult;
  stop*=mult; if ( (stop<0) || (stop+skip>mnpts) ) stop=mnpts-skip;
  
  for(t=skip;t<skip+stop;t++)
    {
      if ( ((int)(t+phase-skip))%((int)(off+on)) >= off )
	orig_X[orig_ev*mnpts+t]=1.0;
      else
	orig_X[orig_ev*mnpts+t]=0.0;
    }
}
break;

/* }}} */
    case 1:
      /* {{{ sinusoid */

{
  float skip, period, phase, stop;

  sprintf(key,"fmri(skip%d)",orig_ev+1);   skip=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(period%d)",orig_ev+1); period=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(phase%d)",orig_ev+1);  phase=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(stop%d)",orig_ev+1);   stop=atof(find_line(fn, key, fl));

  skip*=mult; period*=0.5*mult/M_PI; phase*=mult;
  stop*=mult; if ( (stop<0) || (stop+skip>mnpts) ) stop=mnpts-skip;

  for(t=skip;t<skip+stop;t++)
    orig_X[orig_ev*mnpts+t] = 0.5 * ( 1.0 + sin( M_PI + (t+phase-skip) / period ) );
}
break;

/* }}} */
    case 2:
      /* {{{ custom single column */

{
  FILE *ifp2;

  sprintf(key,"fmri(custom%d)",orig_ev+1); strcpy(filename,(find_line(fn, key, fl)));

  if ((ifp2=fopen(filename,"rb"))==NULL)
    {
      printf("Can't open %s for reading\n",filename);
      exit(1);
    }

  for(i=0,t=0; t<npts; t++)
    {
      float tmpf;
	  
      if (fscanf(ifp2,"%f",&tmpf)!=1)
	{
	  printf("Not enough data in %s\n",filename);
	  exit(1);
	}

      for(; i<trmult*(t+1); i++)
	orig_X[orig_ev*mnpts+i]=tmpf;
    }

  fclose(ifp2);
}
break;

/* }}} */
    case 3:
      /* {{{ custom 3 columns */

{
  FILE *ifp2;
  float start, stop, value;
  int   success=0;

  sprintf(key,"fmri(custom%d)",orig_ev+1); strcpy(filename,(find_line(fn, key, fl)));

  if ((ifp2=fopen(filename,"rb"))==NULL)
    {
      printf("Can't open %s for reading\n",filename);
      exit(1);
    }

  negpts[orig_ev] = NEGSECS*mult;

  while( fscanf(ifp2,"%f %f %f",&start,&stop,&value) == 3 )
    {
      start=start*mult+negpts[orig_ev]; stop=start+stop*mult;

      if ( (stop>0) && (stop-start<1) ) stop=start+1.1; /* if very short stim time make sure it happens */

      if (start<0)     start=0;    /* don't enter stuff before t=0 */
      if (start>mnpts) stop=0;     /* don't do anything */
      if (stop>mnpts)  stop=mnpts; /* don't overrun */

      for(t=start; t<stop; t++)
	{
	  orig_X[orig_ev*mnpts+t]=value;
	  success=1;
	}
    }

  fclose(ifp2);

  if ( success==0 )
    {
      printf("No valid\n[onset duration strength]\ntriplets found in %s\n",filename);
      exit(1);
    }
}
break;

/* }}} */
    }

  if (shape[orig_ev]>1)
    /* {{{ demean */

{
  double Xmean=0;
  
  for(t=0;t<npts*trmult+negpts[orig_ev];t++)
    Xmean += orig_X[orig_ev*mnpts+t];
  
  Xmean /= npts*trmult;
  
  for(t=0;t<npts*trmult+negpts[orig_ev];t++)
    orig_X[orig_ev*mnpts+t] -= Xmean;
}

/* }}} */

  /* {{{ triggers */

#define TRIGGER_THRESH 0.001

{
  float previous=0;
  int trigger_count=0;
  
  for(t=0;t<npts*trmult;t++)
    {
      if ( (orig_X[orig_ev*mnpts+t+negpts[orig_ev]]>TRIGGER_THRESH) && (previous<=TRIGGER_THRESH) && (trigger_count<npts*2) )
	{
	  triggers[trigger_count*orig_evs+orig_ev]=t/trmult - 0.5; /* the 0.5 because sampling is halfway through TR */
	  trigger_count++;
	}

      if ( (orig_X[orig_ev*mnpts+t+negpts[orig_ev]]<=TRIGGER_THRESH) && (previous>TRIGGER_THRESH) && (trigger_count<npts*2) )
	{
	  triggers[trigger_count*orig_evs+orig_ev]=t/trmult - 0.5; /* the 0.5 because sampling is halfway through TR */
	  trigger_count++;
	}

      previous=orig_X[orig_ev*mnpts+t+negpts[orig_ev]];
    }
}

/* }}} */
}

/* }}} */
  /* {{{ convolve and resample down in time */

for(orig_ev=real_ev=0; orig_ev<orig_evs; orig_ev++, real_ev++)
  if (shape[orig_ev]<4)
    {
      int convolve, fw=0;
      float *oX=NULL, *cX=NULL;

      sprintf(key,"fmri(convolve%d)",orig_ev+1); convolve=atoi(find_line(fn, key, fl));

      if (convolve>0)
	{
	  int convolve_phase;
	  sprintf(key,"fmri(convolve_phase%d)",orig_ev+1);
	  convolve_phase=(int)(atof(find_line(fn, key, fl)) / DT);

	  switch (convolve)
	    {
	    case 1:
	      /* {{{ Gaussian */

{
  float sigma, delay, tmpf;

  sprintf(key,"fmri(gausssigma%d)",orig_ev+1); sigma=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(gaussdelay%d)",orig_ev+1); delay=atof(find_line(fn, key, fl));

  sigma /= DT;
  delay /= DT;

  fw = delay + sigma*5;
  maxconvwin=MAX(fw,maxconvwin);
  cX=(float *)malloc(sizeof(float)*fw);

  for(i=0; i<fw; i++)
    {
      tmpf=(((float)i)-delay)/sigma;
      tmpf=exp(-0.5*tmpf*tmpf);
      if (i<1/DT) tmpf *= exp( -pow( 1-0.5*(((float)i)*DT) , 6 ) ); /* between 0 and 1s, window start of kernel */
      cX[i]=tmpf;
    }

  renorm_kernel(cX,fw);
  oX=do_convolve(&orig_X[orig_ev*mnpts],cX,mnpts,fw,convolve_phase);
  do_resample(oX, &real_X[real_ev*npts], npts, trmult, negpts[orig_ev]);
}
break;

/* }}} */
	    case 2:
	      /* {{{ Gamma */

{
  float sigma, delay, alpha, beta;

  sprintf(key,"fmri(gammasigma%d)",orig_ev+1); sigma=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(gammadelay%d)",orig_ev+1); delay=atof(find_line(fn, key, fl));

  sigma /= DT;
  delay /= DT;

  alpha = delay*delay / (sigma*sigma);
  beta  = sigma*sigma / delay;

  fw = delay + sigma*5;
  maxconvwin=MAX(fw,maxconvwin);
  cX=(float *)malloc(sizeof(float)*fw);

  for(i=0; i<fw; i++)
    {
      cX[i]=gammpdf(alpha,1/beta,(float)i);
      /*      if (i>trmult*2) cX[i]=0;*/
    }

  renorm_kernel(cX,fw);
  oX=do_convolve(&orig_X[orig_ev*mnpts],cX,mnpts,fw,convolve_phase);
  do_resample(oX, &real_X[real_ev*npts], npts, trmult, negpts[orig_ev]);
}
break;

/* }}} */
	    case 3:
	      /* {{{ SPM99 HRF */

{
  float sigma1=2.449/DT, delay1=6/DT, sigma2=4/DT, delay2=16/DT, ratio=6,
    alpha1 = delay1*delay1 / (sigma1*sigma1),
    beta1  = sigma1*sigma1 / delay1,
    alpha2 = delay2*delay2 / (sigma2*sigma2),
    beta2  = sigma2*sigma2 / delay2;

  fw = delay2 + sigma2*5;
  maxconvwin=MAX(fw,maxconvwin);
  cX=(float *)malloc(sizeof(float)*fw);

  for(i=0; i<fw; i++)
    cX[i]= gammpdf(alpha1,1/beta1,(float)i) - gammpdf(alpha2,1/beta2,(float)i)/ratio;

  renorm_kernel(cX,fw);
  oX=do_convolve(&orig_X[orig_ev*mnpts],cX,mnpts,fw,convolve_phase);
  do_resample(oX, &real_X[real_ev*npts], npts, trmult, negpts[orig_ev]);
}
break;

/* }}} */
	    case 4:
	      /* {{{ Gamma basis functions */

#define WINDOW_FRAC 0.25

{
  int fnumber, fnum;
  float window, sigma, delay, alpha, beta;
	
  sprintf(key,"fmri(basisfnum%d)",orig_ev+1);   fnumber=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(basisfwidth%d)",orig_ev+1); window=atof(find_line(fn, key, fl));

  orig_ev_nreal[orig_ev]+=fnumber-1;
  /*  real_evs+=fnumber-1;*/
  /*real_X = realloc(real_X,sizeof(float)*real_evs*npts);*/

  fw = window / DT;
  maxconvwin=MAX(fw,maxconvwin);
  cX=(float *)malloc(sizeof(float)*fw);

  delay = fw/2;

  for(fnum=1; fnum<=fnumber; fnum++)
    {
      sigma = delay/2;
      alpha = delay*delay / (sigma*sigma);
      beta  = sigma*sigma / delay;

      for(i=0; i<fw; i++)
	{
	  cX[i]=gammpdf(alpha,1/beta,(float)i);
	  if (i>(1-WINDOW_FRAC)*fw)
	    cX[i] *= 1-(i-(1-WINDOW_FRAC)*fw)/(WINDOW_FRAC*fw);
	}

      renorm_kernel(cX,fw);
      oX=do_convolve(&orig_X[orig_ev*mnpts],cX,mnpts,fw,convolve_phase);
      do_resample(oX, &real_X[real_ev*npts], npts, trmult, negpts[orig_ev]);

      delay *= 0.5;
      real_ev++;
    }

  real_ev--;
}
break;

/* }}} */
	    case 5:
	      /* {{{ Sine basis functions */

{
  int fnumber, fnum;
  float window;
	
  sprintf(key,"fmri(basisfnum%d)",orig_ev+1);   fnumber=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(basisfwidth%d)",orig_ev+1); window=atof(find_line(fn, key, fl));

  orig_ev_nreal[orig_ev]+=fnumber-1;
  /*  real_evs+=fnumber-1;*/
  /*real_X = realloc(real_X,sizeof(float)*real_evs*npts);*/

  fw = window / DT;
  maxconvwin=MAX(fw,maxconvwin);
  cX=(float *)malloc(sizeof(float)*fw);

  for(fnum=1; fnum<=fnumber; fnum++)
    {
      for(i=0; i<fw; i++)
	cX[i]=sin(M_PI*i*fnum/fw);

      oX=do_convolve(&orig_X[orig_ev*mnpts],cX,mnpts,fw,convolve_phase);
      do_resample(oX, &real_X[real_ev*npts], npts, trmult, negpts[orig_ev]);

      real_ev++;
    }

  real_ev--;
}
break;

/* }}} */
	    case 6:
	      /* {{{ FIR basis functions */

{
  int fnumber, fnum;
  float window;
	
  sprintf(key,"fmri(basisfnum%d)",orig_ev+1);   fnumber=atof(find_line(fn, key, fl));
  sprintf(key,"fmri(basisfwidth%d)",orig_ev+1); window=atof(find_line(fn, key, fl));

  orig_ev_nreal[orig_ev]+=fnumber-1;
  /*  real_evs+=fnumber-1;*/
  /*real_X = realloc(real_X,sizeof(float)*real_evs*npts);*/

  fw = window / DT;
  maxconvwin=MAX(fw,maxconvwin);
  cX=(float *)malloc(sizeof(float)*fw);

  for(fnum=0; fnum<fnumber; fnum++)
    {
      for(i=0; i<fw; i++)
	if ( (fw*fnum/fnumber<=i) && (i<fw*(fnum+1)/fnumber) )
	  cX[i]=1;
	else
	  cX[i]=0;

      oX=do_convolve(&orig_X[orig_ev*mnpts],cX,mnpts,fw,convolve_phase);
      do_resample(oX, &real_X[real_ev*npts], npts, trmult, negpts[orig_ev]);

      real_ev++;
    }

  real_ev--;
}
break;

/* }}} */
	    }
	}
      else
	{
	  do_resample(&orig_X[orig_ev*mnpts], &real_X[real_ev*npts], npts, trmult, negpts[orig_ev]);
	  if ( shape[orig_ev]==1 )
	    /* {{{ create sinusoid harmonics */

/* this is treated like a convolution to keep the structure simple -
thus the file key value searching is just a duplicate of that from the
principal sinusoid EV already created */

{
  int nharmonics;

  sprintf(key,"fmri(nharmonics%d)",orig_ev+1); nharmonics=atoi(find_line(fn, key, fl));

  if (nharmonics>0)
    {
      float skip, period, phase, stop;
      int harm;

      sprintf(key,"fmri(skip%d)",orig_ev+1);       skip=atof(find_line(fn, key, fl));
      sprintf(key,"fmri(period%d)",orig_ev+1);     period=atof(find_line(fn, key, fl));
      sprintf(key,"fmri(phase%d)",orig_ev+1);      phase=atof(find_line(fn, key, fl));
      sprintf(key,"fmri(stop%d)",orig_ev+1);       stop=atof(find_line(fn, key, fl));

      skip*=mult; period*=0.5*mult/M_PI; phase*=mult;
      stop*=mult; if ( (stop<0) || (stop+skip>mnpts) ) stop=mnpts-skip;

      orig_ev_nreal[orig_ev]+=nharmonics;

      for(harm=1; harm<=nharmonics; harm++)
	{
	  real_ev++;

	  for(t=skip;t<skip+stop;t++)
	    orig_X[orig_ev*mnpts+t] = 0.5 * ( 1.0 + sin( M_PI + (harm+1)*(t+phase-skip) / period ) );
	
	  do_resample(&orig_X[orig_ev*mnpts], &real_X[real_ev*npts], npts, trmult, negpts[orig_ev]);
	}
    }
}

/* }}} */
	}
    }

/* }}} */
  /* {{{ temporal filtering */

temphp_yn = atoi(find_line(fn, "fmri(temphp_yn)", fl));
templp_yn = atoi(find_line(fn, "fmri(templp_yn)", fl));
nltffwhm  = atof(find_line(fn, "fmri(paradigm_hp)", fl));

if ( (templp_yn) || (temphp_yn) )
  for(orig_ev=real_ev=0; orig_ev<orig_evs; orig_ev++)
  {
    sprintf(key,"fmri(tempfilt_yn%d)",orig_ev+1); tempfiltyn=atoi(find_line(fn, key, fl));

    for(i=0; i<orig_ev_nreal[orig_ev]; i++, real_ev++)
      if (tempfiltyn)
	{
	  image_struct im;
	  double hp_sigma=-1, lp_sigma=-1;
	  
	  im.x=im.y=im.z=1;
	  im.t=npts;
	  im.i=(float*)malloc(sizeof(float)*npts);

	  im.dt=16; im.dtmin=-1e10; im.dtmax=1e10;
      
	  if (templp_yn) lp_sigma=2.8/tr;
	  if (temphp_yn) hp_sigma=0.5*nltffwhm/tr;
	  /*printf("%f %f\n",hp_sigma,lp_sigma);*/

	  for(t=0;t<npts;t++)
	    im.i[t] = real_X[real_ev*npts+t];

	  tempfilt(im, hp_sigma, lp_sigma, 0);
	  for(t=0;t<npts;t++)
	    real_X[real_ev*npts+t] = im.i[t];
	}
  }

/* }}} */
  /* {{{ demean */

for(real_ev=0; real_ev<real_evs; real_ev++)
{
  double Xmean=0;
  
  for(t=0;t<npts;t++)
    Xmean += real_X[real_ev*npts+t];
  
  Xmean /= npts;
  
  for(t=0;t<npts;t++)
    real_X[real_ev*npts+t] -= Xmean;
}

/* }}} */
  /* {{{ orthogonalisation */

for(orig_ev=real_ev=0; orig_ev<orig_evs; orig_ev++)
{
  int tmp, tmp_real, *ortho=malloc(sizeof(int)*real_evs), orthos, northos, doortho, tmptmp;
  double *W, *Z;

  sprintf(key,"fmri(ortho%d.0)",orig_ev+1);  doortho=atoi(find_line(fn, key, fl));

  for(tmptmp=0; tmptmp<orig_ev_nreal[orig_ev]; tmptmp++, real_ev++)
    {
      if (doortho)
	{
	  /* {{{ find out which EVs we will orthogonalise WRT, and load them into W */

  for(tmp=northos=0; tmp<orig_evs; tmp++)
    {
      sprintf(key,"fmri(ortho%d.%d)",orig_ev+1,tmp+1); ortho[tmp]=atoi(find_line(fn, key, fl));
      if (ortho[tmp])
	northos += orig_ev_nreal[tmp];
    }

  W=calloc((npts+northos)*northos,sizeof(double));
  Z=calloc(northos,sizeof(double));

  for(tmp=orthos=tmp_real=0; tmp<orig_evs; tmp++)
     for(i=0; i<orig_ev_nreal[tmp]; i++, tmp_real++)
     {
       if (ortho[tmp])
	 {
	   for(t=0;t<npts;t++)
	     W[t*northos+orthos]=real_X[tmp_real*npts+t];
	   orthos++;
	 }
     }

/* }}} */
	  /* {{{ run SVD */

if (orthos>1)
     SVD(W,Z,npts,northos);

/* }}} */
	  /* {{{ orthogonalise current EV WRT SVD output */

for(tmp=0; tmp<northos; tmp++)
{
  float AdotA=0, BdotA=0;
  
  for(t=0;t<npts;t++)
    {
      AdotA += W[t*northos+tmp]*W[t*northos+tmp];
      BdotA += real_X[real_ev*npts+t]*W[t*northos+tmp];
    }
  
  for(t=0;t<npts;t++)
    real_X[real_ev*npts+t] -= BdotA*W[t*northos+tmp]/AdotA;
}

/* }}} */
	}
    }
}

/* }}} */
  /* {{{ interactions */

for(orig_ev=real_ev=0; orig_ev<orig_evs; orig_ev++)
{
  if (shape[orig_ev]==4)
    {
      int tmp, tmp_real;

      for(t=0;t<npts;t++)
	real_X[real_ev*npts+t]=1.0;
  
      for(tmp=tmp_real=0; tmp<orig_ev; tmp++)
	{
	  sprintf(key,"fmri(interactions%d.%d",orig_ev+1,tmp+1);
	  if ( atoi(find_line(fn, key, fl)) )
	    {
	      float themin=real_X[tmp_real*npts];

	      for(t=0;t<npts;t++)
		if (real_X[tmp_real*npts+t]<themin)
		  themin=real_X[tmp_real*npts+t];

	      for(t=0;t<npts;t++)
		{
		  real_X[real_ev*npts+t] *= real_X[tmp_real*npts+t]-themin;
		}
	    }
	  tmp_real+=orig_ev_nreal[tmp];
	}
    }
  real_ev+=orig_ev_nreal[orig_ev];
}

/* }}} */
  /* {{{ add temporal derivs */

for(orig_ev=real_ev=0; orig_ev<orig_evs; orig_ev++)
{
  sprintf(key,"fmri(deriv_yn%d)",orig_ev+1);

  if ( atoi(find_line(fn, key, fl)) )
    {
      orig_ev_nreal[orig_ev]++;

      /*      real_evs++;*/
      /*      real_X = realloc(real_X,sizeof(float)*real_evs*npts);*/
      if (orig_ev<orig_evs-1)
	memmove(&real_X[(real_ev+2)*npts],&real_X[(real_ev+1)*npts],sizeof(float)*npts*(real_evs-2-real_ev));

      /* do end points differently */
      real_X[(real_ev+1)*npts]        = real_X[real_ev*npts+1]      - real_X[real_ev*npts];
      real_X[(real_ev+1)*npts+npts-1] = real_X[real_ev*npts+npts-1] - real_X[real_ev*npts+npts-2];

      for(t=1;t<npts-1;t++)
	real_X[(real_ev+1)*npts+t] = (real_X[real_ev*npts+t+1] - real_X[real_ev*npts+t-1])*0.5;
    }

  real_ev+=orig_ev_nreal[orig_ev];
}

/* }}} */
  /* {{{ demean */

for(real_ev=0; real_ev<real_evs; real_ev++)
{
  double Xmean=0;
  
  for(t=0;t<npts;t++)
    Xmean += real_X[real_ev*npts+t];
  
  Xmean /= npts;
  
  for(t=0;t<npts;t++)
    real_X[real_ev*npts+t] -= Xmean;
}

/* }}} */
  /* {{{ COMMENT add final (constant) column */

#ifdef FoldingComment

if ( atoi(find_line(fn, "fmri(constcol)", fl)) )
{
  for(t=0;t<npts;t++)
    real_X[real_ev*npts+t]=1.0;

  origevs[Xev]=0;
  Xev++;
}

#endif

/* }}} */
} else {
  /* {{{ group level design */

for(real_ev=0; real_ev<real_evs; real_ev++)
  for(t=0;t<npts;t++)
    {
      sprintf(key,"fmri(evg%d.%d)",t+1,real_ev+1);
      real_X[real_ev*npts+t]=atof(find_line(fn, key, fl));
    }

/* }}} */
}

/* }}} */
  /* {{{ write matrix */

sprintf(filename,"%s.mat",argv[1]);
if ((ofp=fopen(filename,"wb"))==NULL)
{
  printf("Can't open %s for writing\n",filename);
  exit(1);
}

fprintf(ofp,"%%! VEST-Waveform File\n");
fprintf(ofp,"/NumWaves	%d\n",real_evs);
fprintf(ofp,"/NumPoints	%d\n",npts);
fprintf(ofp,"/Skip		0\n");

for(real_ev=0; real_ev<real_evs; real_ev++)
  fprintf(ofp,"/Custom	0 %s\n",argv[1]); /* argv[1] is a dummy entry - the values are really used from this file itself */

fprintf(ofp,"\n/Matrix\n");

for(t=0; t<npts; t++)
{
  for(real_ev=0; real_ev<real_evs; real_ev++)
    fprintf(ofp,"%e	",real_X[real_ev*npts+t]);
  fprintf(ofp,"\n");
}

fclose(ofp);

/* }}} */
  /* {{{ write triggers */

if (level==1)
{
  sprintf(filename,"%s.trg",argv[1]);
  if ((ofp=fopen(filename,"wb"))==NULL)
    {
      printf("Can't open %s for writing\n",filename);
      exit(1);
    }

  for(orig_ev=0; orig_ev<orig_evs; orig_ev++)
    {
      for(i=0;i<orig_ev_nreal[orig_ev];i++)
	{
	  if (shape[orig_ev]!=4)
	    {
	      float *diffs_list=malloc(sizeof(float)*2*npts);
	      int j;
	      
	      for(j=0;triggers[j*orig_evs+orig_ev]>-1e6;j+=2)
		{
		  fprintf(ofp,"%e ",triggers[j*orig_evs+orig_ev]);
		  if (triggers[(j+1)*orig_evs+orig_ev]>-1e6)
		    diffs_list[j/2]=triggers[(j+1)*orig_evs+orig_ev]-triggers[j*orig_evs+orig_ev];
		}
	      
	      fprintf(ofp,"%f\n",median(0.5,diffs_list,j/2) + maxconvwin/trmult);
	      free(diffs_list);
	    }
	  else
	    {
	      fprintf(ofp,"0\n");
	    }
	}
    }

  fclose(ofp);
}

/* }}} */
  /* {{{ contrasts */

C=(float *)malloc(sizeof(float)*real_evs*ncon);

for(con=0; con<ncon; con++)
  for(real_ev=0; real_ev<real_evs; real_ev++)
  {
    sprintf(key,"fmri(con_real%d.%d)",con+1,real_ev+1);
    C[con*real_evs+real_ev]=atof(find_line(fn, key, fl));
  }

sprintf(filename,"%s.con",argv[1]);
if ((ofp=fopen(filename,"wb"))==NULL)
{
  printf("Can't open %s for writing\n",filename);
  exit(1);
}

fprintf(ofp,"%%! VEST-Waveform File\n");
fprintf(ofp,"/NumWaves	%d\n",real_evs);
fprintf(ofp,"/NumContrasts	%d\n",ncon);
fprintf(ofp,"\n/Matrix\n");

for(con=0; con<ncon; con++)
{
  for(real_ev=0; real_ev<real_evs; real_ev++)
    fprintf(ofp,"%e ",C[con*real_evs+real_ev]);
  fprintf(ofp,"\n");
}

fclose(ofp);

/* }}} */
  /* {{{ ftests */

if (nftests>0)
{
  F=malloc(sizeof(int)*ncon*nftests);
  
  for(f=0; f<nftests; f++)
    for(con=0; con<ncon; con++)
      {
	sprintf(key,"fmri(ftest_real%d.%d)",f+1,con+1); F[f*ncon+con]=atoi(find_line(fn, key, fl));
      }

  sprintf(filename,"%s.fts",argv[1]);
  if ((ofp=fopen(filename,"wb"))==NULL)
    {
      printf("Can't open %s for writing\n",filename);
      exit(1);
    }

  fprintf(ofp,"%%! VEST-Waveform File\n");
  fprintf(ofp,"/NumWaves	%d\n",ncon);
  fprintf(ofp,"/NumContrasts	%d\n",nftests);
  fprintf(ofp,"\n/Matrix\n");

  for(f=0; f<nftests; f++)
    {
      for(con=0; con<ncon; con++)
	fprintf(ofp,"%d ",F[f*ncon+con]);
      fprintf(ofp,"\n");
    }

  fclose(ofp);
}

/* }}} */
  /* {{{ second-level group memberships */

if (level==2)
{
  G=malloc(sizeof(int)*npts);

  for(t=0;t<npts;t++)
    {
      sprintf(key,"fmri(groupmem.%d)",t+1);
      G[t]=atoi(find_line(fn, key, fl));
    }

  sprintf(filename,"%s.grp",argv[1]);
  if ((ofp=fopen(filename,"wb"))==NULL)
    {
      printf("Can't open %s for writing\n",filename);
      exit(1);
    }
  
  fprintf(ofp,"%%! VEST-Waveform File\n");
  fprintf(ofp,"/NumWaves	1\n");
  fprintf(ofp,"/NumPoints	%d\n",npts);
  fprintf(ofp,"\n/Matrix\n");

  for(t=0;t<npts;t++)
    fprintf(ofp,"%d\n",G[t]);

  fclose(ofp);
}

/* }}} */
  /* {{{ svd to check rank of matrix */

if (real_evs>1)
{
  W=(double *)malloc(sizeof(double)*(npts+real_evs)*real_evs);
  Z=(double *)calloc(real_evs,sizeof(double));
    
  for(real_ev=0; real_ev<real_evs; real_ev++)
    for(t=0;t<npts;t++)
      W[t*real_evs+real_ev]=real_X[real_ev*npts+t];

  SVD(W,Z,npts,real_evs);
  for(real_ev=0; real_ev<real_evs; real_ev++)
    Z[real_ev]=sqrt(Z[real_ev]);
  minl=maxl=Z[0];
  for(real_ev=0; real_ev<real_evs; real_ev++)
    {
      /*printf("%e\n",Z[real_ev]);*/
      if (Z[real_ev]<minl) minl=Z[real_ev];
      if (Z[real_ev]>maxl) maxl=Z[real_ev];
    }
  condition=minl/maxl;

  /*printf("condition = %e\n",condition);*/

  if (condition<MINCONDITION)
    printf("At least one EV is (close to) a linear combination of the others. You should probably alter your design.\n\n(Design matrix is rank deficient - ratio of min:max eigenvalues in SVD of matrix is %e)\n",condition);
}

/* }}} */
  /* {{{ write autocov matrix */

if (real_evs>1)
{
  /* {{{ setup vars */

  FILE *ofp2;
  unsigned char *r,*g,*b;
  int border=5, mag, size, xsize, ysize, x, y, evx, evy;
  float *cov=malloc(real_evs*real_evs*sizeof(float)), *covnorm=malloc(real_evs*sizeof(float));

  mag  = MIN( MAX( 1 , 300/real_evs  ) , 50);

  size  = real_evs*mag;
  xsize = size*2 + border*3;
  ysize = size   + border*2;

/* }}} */
  /* {{{ malloc images and fill in background */

r=(unsigned char *)malloc(xsize*ysize);
g=(unsigned char *)malloc(xsize*ysize);
b=(unsigned char *)malloc(xsize*ysize);

memset((void *)r,(unsigned char)180,xsize*ysize);
memset((void *)g,(unsigned char)215,xsize*ysize);
memset((void *)b,(unsigned char)255,xsize*ysize);

/* }}} */
  /* {{{ cov matrix orig */

for(evx=0; evx<real_evs; evx++)
  for(evy=0; evy<real_evs; evy++)
  {
    cov[evy*real_evs+evx]=0;
    for(t=0;t<npts;t++)
      cov[evy*real_evs+evx]+=real_X[evx*npts+t]*real_X[evy*npts+t];
    if (evx==evy)
      covnorm[evx]=cov[evx*real_evs+evx];
  }

for(evx=0; evx<real_evs; evx++)
  for(evy=0; evy<real_evs; evy++)
  {
    tmpf=covnorm[evy]*covnorm[evx];
    if (tmpf<=0)
      cov[evy*real_evs+evx]=0;
    else
      cov[evy*real_evs+evx]=ABS(cov[evy*real_evs+evx])/sqrt(tmpf);
  }

for(evx=0; evx<real_evs; evx++)
  for(evy=0; evy<real_evs; evy++)
    for(y=0; y<mag; y++)
      for(x=0; x<mag; x++)
        r[(border+evy*mag+y)*xsize+border+evx*mag+x] =
        g[(border+evy*mag+y)*xsize+border+evx*mag+x] =
        b[(border+evy*mag+y)*xsize+border+evx*mag+x] = 255*cov[evy*real_evs+evx];

/* }}} */
  /* {{{ cov matrix SVD */

tmpf=0;
for(evx=0; evx<real_evs; evx++)
  for(evy=0; evy<real_evs; evy++)
  {
    if (evx!=evy)
      cov[evy*real_evs+evx]=0;
    else
      {
	cov[evy*real_evs+evx]=Z[evx];
	if (cov[evy*real_evs+evx]>tmpf)
	  tmpf=cov[evy*real_evs+evx];
      }
  }

for(evx=0; evx<real_evs; evx++)
  cov[evx*real_evs+evx]/=tmpf;

for(evx=0; evx<real_evs; evx++)
  for(evy=0; evy<real_evs; evy++)
    for(y=0; y<mag; y++)
      for(x=0; x<mag; x++)
        r[(border+evy*mag+y)*xsize+2*border+size+evx*mag+x] =
        g[(border+evy*mag+y)*xsize+2*border+size+evx*mag+x] =
        b[(border+evy*mag+y)*xsize+2*border+size+evx*mag+x] = 255*cov[evy*real_evs+evx];

/* }}} */
  /* {{{ output image */

sprintf(filename,"%s_cov.ppm",argv[1]);
if ((ofp2=fopen(filename,"wb"))==NULL)
{
  printf("Can't open %s for writing\n",filename);
  exit(1);
}

fprintf(ofp2,"P6\n");
fprintf(ofp2,"%d %d\n",xsize,ysize);
fprintf(ofp2,"255\n");

for(y=0; y<ysize; y++)
     for(x=0; x<xsize; x++)
{
  fwrite(&r[y*xsize+x],1,1,ofp2);
  fwrite(&g[y*xsize+x],1,1,ofp2);
  fwrite(&b[y*xsize+x],1,1,ofp2);
}

fclose(ofp2);

sprintf(filename,"%s/bin/convert %s_cov.ppm %s_cov.gif",FSLDIR,argv[1],argv[1]);
system(filename);

/* }}} */
}

/* }}} */
  /* {{{ write image preview (corrupts X[]) */

{
  /* {{{ setup vars */

FILE *ofp2;
unsigned char *r,*g,*b;
int xmag, ymag, border=5, temp=15, fsize=11, xsize, ysize, xx, yy, x, y, name_length=temp;

/* how much space for contrast names? */
for(yy=0; yy<ncon; yy++)
{
  sprintf(key,"fmri(conname_real.%d)",yy+1);
  strcpy(the_string,find_line(fn, key, fl));
  xx=(strlen(the_string)+4) * FONT_WIDTH;
  if (xx>name_length) name_length=xx;
}

xmag  = MIN( MAX( 1 , 600/real_evs  ) , 50);
ymag  = MIN( MAX( 1 , 400/npts ) , 20);

xsize=real_evs*xmag + border*(real_evs+3+nftests+(nftests>0)) + name_length + nftests*fsize;
ysize=npts*ymag + ncon*FONT_HEIGHT + border*(3+ncon);

/* }}} */
  /* {{{ reset X[] range (but don't change offset) */

for(real_ev=0; real_ev<real_evs; real_ev++)
{
  float absmax=0;

  for(t=0;t<npts;t++)
    if ( fabs(real_X[real_ev*npts+t]) > absmax )
      absmax=fabs(real_X[real_ev*npts+t]);

  if (absmax>0)
    for(t=0;t<npts;t++)
      real_X[real_ev*npts+t] /= absmax;
}

/* }}} */
  /* {{{ malloc images and fill in background */

r=(unsigned char *)malloc(xsize*ysize);
g=(unsigned char *)malloc(xsize*ysize);
b=(unsigned char *)malloc(xsize*ysize);

memset((void *)r,(unsigned char)180,xsize*ysize);
memset((void *)g,(unsigned char)215,xsize*ysize);
memset((void *)b,(unsigned char)255,xsize*ysize);

/* }}} */
  /* {{{ time */

if (level==1)
for(yy=0; yy<npts; yy++)
  for(y=0; y<ymag; y++)
    for(x=0; x<temp; x++)
    {
      int intensity=64, rintensity;

      if (yy%10==0) intensity=255;

      rintensity=intensity;
      if (abs(yy-npts/2)<=(0.5*nltffwhm/tr))
	{
	  if (x>1&&x<temp-2)
	    rintensity=intensity=0;
	  if (x>3&&x<temp-4)
	    {
	      rintensity=255;
	      intensity=0;
	    }
	}

      r[(border+yy*ymag+y)*xsize+border+x]=rintensity;
      g[(border+yy*ymag+y)*xsize+border+x]=intensity;
      b[(border+yy*ymag+y)*xsize+border+x]=intensity;
    }

/* }}} */
  /* {{{ DM: grey */

  for(yy=0; yy<npts; yy++)
    for(xx=0; xx<real_evs; xx++)
      for(y=0; y<ymag; y++)
	for(x=0; x<xmag; x++)
	  {
	    float XX = real_X[xx*npts+yy];

	    r[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+name_length]=(XX * 100.0)+128;
	    g[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+name_length]=(XX * 100.0)+128;
	    b[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+name_length]=(XX * 100.0)+128;
	  }

/* }}} */
  /* {{{ DM: red */

  for(yy=0; yy<npts; yy++)
    for(xx=0; xx<real_evs; xx++)
      for(y=0; y<ymag; y++)
	{
	  float XX = real_X[xx*npts+yy];
	  
	  x = (XX*0.8+1.0)*xmag/2;

	  r[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+name_length]=255;
	  g[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+name_length]=0;
	  b[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+name_length]=0;
	  r[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+1+name_length]=255;
	  g[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+1+name_length]=0;
	  b[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+1+name_length]=0;
	  r[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x-1+name_length]=255;
	  g[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x-1+name_length]=0;
	  b[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x-1+name_length]=0;
	  r[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+2+name_length]=0;
	  g[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+2+name_length]=0;
	  b[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x+2+name_length]=0;
	  r[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x-2+name_length]=0;
	  g[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x-2+name_length]=0;
	  b[(border+yy*ymag+y)*xsize+border*(xx+3)+xx*xmag+x-2+name_length]=0;
	}

/* }}} */
  /* {{{ contrasts */

for(yy=0; yy<ncon; yy++)
{
  sprintf(the_string,"C%d",yy+1);
  write_string_rgb(r, g, b, 
		   border,
		   (3+yy)*border+yy*FONT_HEIGHT+npts*ymag,
		   the_string, font_data, 0, 0, 0, xsize, ysize);

  sprintf(key,"fmri(conname_real.%d)",yy+1);
  strcpy(the_string,find_line(fn, key, fl));
  write_string_rgb(r, g, b, 
		   border+4*FONT_WIDTH,
		   (3+yy)*border+yy*FONT_HEIGHT+npts*ymag,
		   the_string, font_data, 150, 50, 50, xsize, ysize);

  for(xx=0; xx<real_evs; xx++)
    {
      if ( C[yy*real_evs+xx] == (int)C[yy*real_evs+xx])
	sprintf(the_string,"%d",(int)C[yy*real_evs+xx]);
      else
	sprintf(the_string,"%.3f",C[yy*real_evs+xx]);
      
      write_string_rgb(r, g, b, 
		       (int)(border*(xx+3)+(xx+0.5)*xmag+name_length-FONT_WIDTH*0.5*strlen(the_string)),
		       (3+yy)*border+yy*FONT_HEIGHT+npts*ymag,
		       the_string, font_data, 0, 0, 0, xsize, ysize);
    }
}

/* }}} */
  /* {{{ ftests */

for(f=0; f<nftests; f++)
{
  write_string_rgb(r, g, b,
		   border*(real_evs+4)+real_evs*xmag+name_length+f*(border+fsize),
		   border-2*FONT_HEIGHT+npts*ymag,
		   "F", font_data, 0, 0, 0, xsize, ysize);

  sprintf(the_string,"%d",f+1);
  write_string_rgb(r, g, b,
		   border*(real_evs+4)+real_evs*xmag+name_length+f*(border+fsize),
		   border-FONT_HEIGHT+npts*ymag,
		   the_string, font_data, 0, 0, 0, xsize, ysize);

  for(con=0; con<ncon; con++)
    {
      for(y=0; y<fsize; y++)
        for(x=0; x<fsize; x++)
	    r[ ((3+con)*border+con*FONT_HEIGHT+npts*ymag+y)*xsize +
	     border*(real_evs+4)+real_evs*xmag+name_length+f*(border+fsize)+x ]=
            g[ ((3+con)*border+con*FONT_HEIGHT+npts*ymag+y)*xsize +
	     border*(real_evs+4)+real_evs*xmag+name_length+f*(border+fsize)+x ]=
            b[ ((3+con)*border+con*FONT_HEIGHT+npts*ymag+y)*xsize +
	     border*(real_evs+4)+real_evs*xmag+name_length+f*(border+fsize)+x ]=0;

      for(y=1; y<fsize-1; y++)
	for(x=1; x<fsize-1; x++)
	  if (F[f*ncon+con])
	    r[ ((3+con)*border+con*FONT_HEIGHT+npts*ymag+y)*xsize +
	     border*(real_evs+4)+real_evs*xmag+name_length+f*(border+fsize)+x ]=175;
	  else
	    {
	      r[ ((3+con)*border+con*FONT_HEIGHT+npts*ymag+y)*xsize +
	       border*(real_evs+4)+real_evs*xmag+name_length+f*(border+fsize)+x ]=180;
	      g[ ((3+con)*border+con*FONT_HEIGHT+npts*ymag+y)*xsize +
	       border*(real_evs+4)+real_evs*xmag+name_length+f*(border+fsize)+x ]=215;
	      b[ ((3+con)*border+con*FONT_HEIGHT+npts*ymag+y)*xsize +
	       border*(real_evs+4)+real_evs*xmag+name_length+f*(border+fsize)+x ]=255;

	    }
    }
}

/* }}} */
  /* {{{ second-level group memberships */

if (level==2)
  for(t=0;t<npts;t++)
    {
      sprintf(the_string,"%d",G[t]);
      write_string_rgb(r, g, b,
		       border,
		       border+t*ymag,
		       the_string, font_data, 0, 0, 0, xsize, ysize);
    }

/* }}} */
  /* {{{ COMMENT FEAT */

#ifdef FoldingComment

sprintf(the_string,"FEAT %s",find_line(fn, "fmri(version)", fl));
write_string_rgb(r, g, b,
		 xsize-1-border-FONT_WIDTH*strlen(the_string),
		 border,
		 the_string, font_data, 255, 255, 0, xsize, ysize);

#endif

/* }}} */
  /* {{{ output image */

sprintf(filename,"%s.ppm",argv[1]);
if ((ofp2=fopen(filename,"wb"))==NULL)
{
  printf("Can't open %s for writing\n",filename);
  exit(1);
}

fprintf(ofp2,"P6\n");
fprintf(ofp2,"%d %d\n",xsize,ysize);
fprintf(ofp2,"255\n");

for(y=0; y<ysize; y++)
     for(x=0; x<xsize; x++)
{
  fwrite(&r[y*xsize+x],1,1,ofp2);
  fwrite(&g[y*xsize+x],1,1,ofp2);
  fwrite(&b[y*xsize+x],1,1,ofp2);
}

fclose(ofp2);

sprintf(filename,"%s/bin/convert %s.ppm %s.gif",FSLDIR,argv[1],argv[1]);
system(filename);

/* }}} */
}

/* }}} */

  return(0);
}

