/*  gsmanager.cc

    Mark Woolrich, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  CCOPYRIGHT  */

#include "gsmanager.h"
#include "Log.h"
#include "miscmaths.h"
#include "miscprob.h"
#include "newimageall.h"
#include "tracer_plus.h"
//#include "mcmc.h"
#include "mcmc_mh.h"
#include "t2z.h"
#include "f2z.h"

using namespace Utilities;
using namespace MISCMATHS;
using namespace NEWIMAGE;

namespace Gs {

  float Gsmanager::mdofls(const ColumnVector& xsq, const float phi, const int P) const
  {    
    Tracer_Plus trace("Gsmanager::mdofls");

    ColumnVector xsqscaled = xsq/phi;

    float vbig=200;
    float vmid=31;
    float vsm=0.1;
    float fbig=mgradpt(vbig,xsqscaled,P);
    float fmid=mgradpt(vmid,xsqscaled,P);
    float fsm=mgradpt(vsm,xsqscaled,P);

    while((vbig-vmid)>0.5 && (vmid-vsm)>0.5)
      {  
	if(sign(fsm)!=sign(fmid))
	  {
	    vbig=vmid;
	    fbig=fmid;
	    vmid=(vmid+vsm)/2.0;
	    fmid=mgradpt(vmid,xsqscaled,P);
	  }
	else if(sign(fmid)!=sign(fbig))
	  {
	    vsm=vmid;
	    fsm=fmid;
	    vmid=(vbig+vmid)/2.0;
	    fmid=mgradpt(vmid,xsqscaled,P);
	  }
	else
	  {
	    vsm=100;
	    vmid=100;
	    vbig=100;
	  }
      }

    return round(vmid);
  }

  float Gsmanager::mgradpt(const float v, const ColumnVector& xsq, const int P) const
  {
    Tracer_Plus trace("Gsmanager::mgradpt");

    int n = xsq.Nrows();

    float sumoveri = 0.0;

    float vplusPon2v2 = (v+P)/(2.0*v);

    for(int i=1; i<=n; i++)
      {
	sumoveri += vplusPon2v2*xsq(i)/(v+xsq(i)) - 0.5*log(1+xsq(i)/v);
      }
    
    return -(n*(0.5*digamma((v+P)/2.0) - 0.5*P/v - 0.5*digamma(v/2.0)) + sumoveri);	     

  }

  float Gsmanager::csevl(const float x, const ColumnVector& cs, const int n) const
  {
    Tracer_Plus trace("Gsmanager::csevl");

    float b0 = 0;
    float b1 = 0;
    float b2 = 0;
    const float twox=2*x;
    
    for(int i=1; i<=n; i++)
      {
	b2=b1;
	b1=b0;
	b0=twox*b1-b2+cs(n+1-i);
      }
    
    return 0.5*(b0-b2);
  }

  float Gsmanager::digamma(const float x) const
  { 
    Tracer_Plus trace("Gsmanager::digamma");

    float y = fabs(x);
    float psi;

    if(y<2.0)
      {
	// do we need to deal with the following case?
	// c psi(x) for -2. .lt. x .lt. 2.

	int n = int(floor(x));
	y = x - n;
	n = n - 1;
	psi = csevl(2*y-1, psics, ntpsi);
	if(n==-1)
	  {
	    psi = psi - 1.0/x;
	  }
      }
    else
      {
	const float aux = csevl(8/(Sqr(y))-1, apsics, ntapsi);
	psi = log(x) - 0.5/x + aux;
      }
    
    return psi;
  }

  void Gsmanager::multitfit(const Matrix& x, ColumnVector& m, SymmetricMatrix& covar, float& v) const
  {
    Tracer_Plus trace("Gsmanager::multitfit");

    int n = x.Ncols();
    int P = x.Nrows();

    v = 10;

    Matrix mx;
    Matrix mn;
    remmean(x,mx,mn,2);
    m = mn.t().AsColumn();

    covar = cov(mx.t());
    float tmp = pow(covar.Determinant(),(1.0/P));    

    covar=covar/tmp;

    // xsq(i) = x(i)'*inv(c)*x(i)
    ColumnVector xsq(n);
    SymmetricMatrix invcovar = covar.i();
    
    for(int i = 1; i <= n; i++)
      {
	xsq(i) = (mx.Column(i).t()*invcovar*mx.Column(i)).AsScalar();
      }

    float phi = tmp;

    for(int i = 1; i <= 20; i++)
      {
	float newphi = 0.0;
	for(int j = 1; j <= n; j++)
	  {
	    float tau = phi*(v+P)/(v*phi+xsq(j));
	    newphi += tau*xsq(j);
	  }
	phi = newphi/float(n*P);

// 	write_ascii_matrix(xsq,'xsq');
// 	OUT(P);
// 	OUT(phi);
	v = mdofls(xsq,phi,P);
//	OUT(v);
      } 

    covar = covar*phi;
  }

  float Gsmanager::logtpdf(const float& v, const ColumnVector& x2, const float& phi) const
  {
    int n = x2.Nrows();

    float sumoverj = 0.0;

    for(int j = 1; j <= n; j++)
      {
	sumoverj += log(1+(x2(j)/phi)/float(v));
      }

    return n*lgam((v+1)/2.0) - n*lgam(v/2.0) - ((v+1)/2.0)*sumoverj - n*0.5*log(M_PI*v);
  }

  void Gsmanager::setup()
  {
    Tracer_Plus trace("Gsmanager::setup");
    
    if(GsOptions::getInstance().debuglevel.value()==2)
      {
	cout << "******************************************" << endl
	     << "SETUP" << endl << "******************************************"
	     << endl;
      }

    // read data
//     read_volume4D(copedata, opts.copefile.value(),volinfo);
//     read_volume4D(varcopedata, opts.varcopefile.value());
    copedata.read(opts.copefile.value());
    varcopedata.read(opts.varcopefile.value());

    // mask:
    read_volume(mask, opts.maskfile.value(),volinfo);

    sparsemask.read(opts.maskfile.value());

    zmask = mask;
    zmask.binarise(1e-16);  // newimage call
    nmaskvoxels = int(zmask.sum());

    sparsemask.threshold(1e-16);

    if(sparsemask.getVolumeSize() != nmaskvoxels) throw Exception("masks not consistent");
    
    // threshold using mask:
    copedata.setPreThresholdPositions(sparsemask.getPreThresholdPositions());
    copedata.thresholdSeries();

    varcopedata.setPreThresholdPositions(sparsemask.getPreThresholdPositions());
    varcopedata.thresholdSeries();

//     if(opts.dofvarcopefile.value() != string(""))
//       read_volume4D(dofvarcopedata, opts.dofvarcopefile.value());
    if(opts.dofvarcopefile.value() != string(""))
      {
	dofvarcopedata.read(opts.dofvarcopefile.value());
	dofvarcopedata.setPreThresholdPositions(sparsemask.getPreThresholdPositions());
	dofvarcopedata.thresholdSeries();
      }
    else
      {
	dofvarcopedata = varcopedata;
	varcopedata = 0;
      }

    if(GsOptions::getInstance().debuglevel.value()==2)
      {
	cout << "******************************************" << endl
	     << "Data Read Complete" << endl << "******************************************"
	     << endl;
      }
        

    // setup design
    design.setup();    
    
    // check design and data compatability
    if(design.getntpts() != copedata.tsize())
      {	
	cout << "design.getntpts()=" << design.getntpts() << endl;
	cout << "copedata.tsize()=" << copedata.tsize() << endl;
	throw Exception("Cope data and design have different numbers of time points");
      }
    if(design.getntpts() != varcopedata.tsize())
      {	
	cout << "dm.ntpts()=" << design.getntpts() << endl;
	cout << "copedata.tsize()=" << varcopedata.tsize() << endl;
	
	throw Exception("Varcope data and design have different numbers of time points");
      }
    
    ngs = design.getngs();
    nevs = design.getnevs();
    ntpts = design.getntpts();
    xsize = mask.xsize();
    ysize = mask.ysize();
    zsize = mask.zsize();
    
    cout << "nevs=" << nevs << endl;
    cout << "ntpts=" << ntpts << endl;
    cout << "ngs=" << ngs << endl;
    cout << "nvoxels=" << nmaskvoxels << endl;

    initialise();

  }

  void Gsmanager::initialise()
  {
    Tracer_Plus trace("Gsmanager::initialise");

    pes.resize(design.getnevs());
    ts.resize(design.getnumtcontrasts());
    tdofs.resize(design.getnumtcontrasts());
    zts.resize(design.getnumtcontrasts());
    tcopes.resize(design.getnumtcontrasts());
    tvarcopes.resize(design.getnumtcontrasts());
    fs.resize(design.getnumfcontrasts());
    fdof1s.resize(design.getnumfcontrasts());
    fdof2s.resize(design.getnumfcontrasts());
    zfs.resize(design.getnumfcontrasts());
       
    beta_b.resize(design.getngs());
    beta_c.resize(design.getngs());
    beta_mean.resize(design.getngs());

    cov_pes.ReSize(nevs*nevs,nmaskvoxels);   
    cov_pes = 0;

    for(int g = 0; g < ngs; g++)
      {
	beta_mean[g].ReSize(nmaskvoxels);		
	beta_mean[g] = 0;
	beta_b[g].ReSize(nmaskvoxels);		
	beta_b[g] = 0;
	beta_c[g].ReSize(nmaskvoxels);		
	beta_c[g] = 0;

      }

    for(int e = 0; e < nevs; e++)
      {
	pes[e].ReSize(nmaskvoxels);
	pes[e] = 0;
      }

    for(int t = 0; t < design.getnumtcontrasts(); t++)
      {
	ts[t].ReSize(nmaskvoxels);
	ts[t] = 0;
	tdofs[t].ReSize(nmaskvoxels);

	tdofs[t] = 0;
	zts[t].ReSize(nmaskvoxels);
	zts[t] = 0;
	tcopes[t].ReSize(nmaskvoxels);
	tcopes[t] = 0;
	tvarcopes[t].ReSize(nmaskvoxels);
	tvarcopes[t] = 0;
      }

    for(int f = 0; f < design.getnumfcontrasts(); f++)
      {
	fs[f].ReSize(nmaskvoxels);
	fs[f] = 0;
	fdof1s[f].ReSize(nmaskvoxels);
	fdof1s[f] = 0;
	fdof2s[f].ReSize(nmaskvoxels);
	fdof2s[f] = 0;
	zfs[f].ReSize(nmaskvoxels);
	zfs[f] = 0;
      }

    ols();    
  } 

  void Gsmanager::run()
  {
    Tracer_Plus trace("Gsmanager::run");

    // MCMC
    vector<volume4D<float> > gamma_samples(nevs);
    vector<volume<float> > gamma_naccepted(nevs);
    vector<volume<float> > gamma_nrejected(nevs);
    vector<volume4D<float> > beta_samples(ngs);
    vector<volume<float> > beta_naccepted(ngs);
    vector<volume<float> > beta_nrejected(ngs);
    vector<volume4D<float> > phi_samples(ntpts);
    vector<volume<float> > phi_naccepted(ntpts);
    vector<volume<float> > phi_nrejected(ntpts);
    vector<volume4D<float> > ss_samples(design.getnumfcontrasts()+1);
    
    volume<float> DIC(xsize,ysize,zsize);
    volume<float> pd(xsize,ysize,zsize);
    DIC = 0;
    pd = 0;

    int nsamples = (opts.njumps.value()-opts.burnin.value())/opts.sampleevery.value();
    cout << "njumps = " << opts.njumps.value() << endl;
    cout << "burnin = " << opts.burnin.value() << endl;
    cout << "sampleevery = " << opts.sampleevery.value() << endl;
    cout << "nsamples = " << nsamples << endl << endl;
        

    if(GsOptions::getInstance().verbose.value())
      {
	for(int e = 0; e < nevs; e++)
	  {	
	    gamma_samples[e].reinitialize(xsize,ysize,zsize,nsamples);
	    gamma_samples[e] = 0;
	    gamma_naccepted[e].reinitialize(xsize,ysize,zsize);
	    gamma_naccepted[e] = 0;
	    gamma_nrejected[e].reinitialize(xsize,ysize,zsize);
	    gamma_nrejected[e] = 0;
	  }

	for(int t = 0; t < ntpts; t++)
	  {
	    phi_samples[t].reinitialize(xsize,ysize,zsize,nsamples);
	    phi_samples[t] = 0;
	    phi_naccepted[t].reinitialize(xsize,ysize,zsize);
	    phi_naccepted[t] = 0;
	    phi_nrejected[t].reinitialize(xsize,ysize,zsize);
	    phi_nrejected[t] = 0;
	  }
	
	for(int g = 0; g < ngs; g++)
	  {
	    beta_samples[g].reinitialize(xsize,ysize,zsize,nsamples);
	    beta_samples[g] = 0;
	    beta_naccepted[g].reinitialize(xsize,ysize,zsize);
	    beta_naccepted[g] = 0;
	    beta_nrejected[g].reinitialize(xsize,ysize,zsize);
	    beta_nrejected[g] = 0;
	  }

	for(int f = 0; f < design.getnumfcontrasts()+1; f++)
	  {
	    ss_samples[f].reinitialize(xsize,ysize,zsize,nsamples);
	    ss_samples[f] = 0;
	  }

      }

    Matrix gamsamples(nevs, nsamples);
    Matrix betasamples(ngs, nsamples);
    Matrix phisamples(ntpts, nsamples);
    ColumnVector likelihood_samples(nsamples);
    vector<ColumnVector> sssamples(design.getnumfcontrasts()+1);

    for(int f = 0; f < design.getnumfcontrasts()+1; f++)
      {
	sssamples[f].ReSize(nsamples);
	sssamples[f] = 0;
      }

    int vox = 0;
    int vox2 = 0;
    int voxout = 0;

    cout << "Metropolis Hasting Sampling" << endl;
    cout << "Number of voxels=" << nmaskvoxels << endl;
    cout << "Percentage done:" << endl;

//     VolumeSeries test = copedata;
//     test = 0;

    for(int x = 0; x < xsize; x++)
      for(int y = 0; y < ysize; y++)
	for(int z = 0; z < zsize; z++)
	  {
	    if(mask(x,y,z))
	      {
		vox++;
		
		//		test.setSeries(copedata.getSeries(vox),vox);

		if(zmask(x,y,z))
		  {
		    vox2++;
		    gamsamples = 0;
		    betasamples = 0;
		    phisamples = 0;

		    if(vox2 > voxout*nmaskvoxels/100.0)
		      {
			//cout<<(voxout+1)<<'%' <<'\r';		
			cout << " " << (voxout+1);
			cout.flush();
			voxout++;
		      }

		    if(GsOptions::getInstance().debuglevel.value()==2)
		      {
			cout << "--------------" << endl;
			cout << "x=" << x << "y=" << y << "z=" << z << endl;
		      }

		    ColumnVector gammamean(nevs);
		    for(int e = 0; e < nevs; e++)
		      {
			gammamean(e+1) = pes[e](vox);
		      }
		    ColumnVector betab(ngs);
		    ColumnVector betac(ngs);
		    for(int g = 0; g < ngs; g++)
		      {
			betab(g+1) = beta_b[g](vox);
			betac(g+1) = beta_c[g](vox);
		      }

		    Mcmc_Mh mcmc_mh(copedata.getSeries(vox), varcopedata.getSeries(vox),dofvarcopedata.getSeries(vox), design, gammamean, cov_pes.getSeries(vox), betab, betac, gamsamples, betasamples, phisamples, likelihood_samples, sssamples, nsamples);
		    mcmc_mh.setup();
		    mcmc_mh.run();

		    for(int e = 0; e < nevs; e++)
		      {
			pes[e](vox) = mean(gamsamples.Row(e+1).t()).AsScalar();
		      }
		    for(int g = 0; g < ngs; g++)
		      {
			beta_mean[g](vox) = mean(betasamples.Row(g+1).t()).AsScalar();
		      }

		    mcmc_contrasts(gamsamples,vox);

		    mcmc_mh.dic(DIC(x,y,z), pd(x,y,z));

		    if(GsOptions::getInstance().verbose.value())
		      {
			for(int e = 0; e < nevs; e++)
			  {
			    gamma_samples[e].setvoxelts(gamsamples.Row(e+1).t(),x,y,z);
			    gamma_naccepted[e](x,y,z) = mcmc_mh.getgamma_naccepted()(e+1);
			    gamma_nrejected[e](x,y,z) = mcmc_mh.getgamma_nrejected()(e+1);
			  }	

			for(int f = 0; f < design.getnumfcontrasts()+1; f++)
			  {
			    ss_samples[f].setvoxelts(sssamples[f],x,y,z);
			  }
		    
			for(int g = 0; g < ngs; g++)
			  {
			    beta_samples[g].setvoxelts(betasamples.Row(g+1).t(),x,y,z);
			    beta_naccepted[g](x,y,z) = mcmc_mh.getbeta_naccepted()(g+1);
			    beta_nrejected[g](x,y,z) = mcmc_mh.getbeta_nrejected()(g+1);
			  }

			for(int t = 0; t < ntpts; t++)
			  {
			    phi_samples[t].setvoxelts(phisamples.Row(t+1).t(),x,y,z);
			    phi_naccepted[t](x,y,z) = mcmc_mh.getphi_naccepted()(t+1);
			    phi_nrejected[t](x,y,z) = mcmc_mh.getphi_nrejected()(t+1);
			  }

		      }
		  }	      	    
	      }
	  }

//     VolumeInfo inf = sparsemask.getInfo();
//     inf.v = test.tsize();
//     test.setInfo(inf);
//     test.setPreThresholdPositions(sparsemask.getPreThresholdPositions());
//     test.unthresholdSeries();
//     test.writeAsFloat("test");
	
    if(GsOptions::getInstance().verbose.value())
      {
	for(int g = 0; g < ngs; g++)
	  {
	    save_volume4D(beta_samples[g], LogSingleton::getInstance().appendDir("beta"+num2str(g+1)+"_samples"));	
	    save_volume(beta_naccepted[g], LogSingleton::getInstance().appendDir("beta"+num2str(g+1)+"_naccepted"));
	    save_volume(beta_nrejected[g], LogSingleton::getInstance().appendDir("beta"+num2str(g+1)+"_nrejected"));	
	  }

	for(int t = 0; t < ntpts; t++)
	  {
	    save_volume4D(phi_samples[t], LogSingleton::getInstance().appendDir("phi"+num2str(t+1)+"_samples"));	
	    save_volume(phi_naccepted[t], LogSingleton::getInstance().appendDir("phi"+num2str(t+1)+"_naccepted"));
	    save_volume(phi_nrejected[t], LogSingleton::getInstance().appendDir("phi"+num2str(t+1)+"_nrejected"));	
	  }
	
	for(int e = 0; e < nevs; e++)
	  {
	    save_volume4D(gamma_samples[e], LogSingleton::getInstance().appendDir("gamma"+num2str(e+1)+"_samples"));
	    save_volume(gamma_naccepted[e], LogSingleton::getInstance().appendDir("gamma"+num2str(e+1)+"_naccepted"));
	    save_volume(gamma_nrejected[e], LogSingleton::getInstance().appendDir("gamma"+num2str(e+1)+"_nrejected"));
	    
	  }
	
// 	for(int f = 0; f < design.getnumfcontrasts()+1; f++)
// 	  {
// 	    save_volume4D(ss_samples[f], LogSingleton::getInstance().appendDir("ss"+num2str(f+1)+"_samples"));
// 	  }

	save_volume(DIC, LogSingleton::getInstance().appendDir("dic"));
	save_volume(pd, LogSingleton::getInstance().appendDir("pd"));
      }
    
    cout << endl;
  }


  void Gsmanager::save()
  {
    Tracer_Plus trace("Gsmanager::save");

    varcopedata.CleanUp();
    dofvarcopedata.CleanUp();

    // need to save residuals:
    VolumeSeries res = copedata;
    res = 0;
    
    const Matrix& dm = design.getdm();
    int vox = 0;
    for(int x = 0; x < xsize; x++)
      for(int y = 0; y < ysize; y++)
	for(int z = 0; z < zsize; z++)
	  {
	    if(mask(x,y,z))
	      {
		vox++;
		ColumnVector petmp(nevs);
		for(int e = 0; e < nevs; e++)
		  {
		    petmp(e+1) = pes[e](vox);
		  }

		res.setSeries(copedata.getSeries(vox)-dm*petmp,vox);
	      }
	  }

    VolumeInfo inf = sparsemask.getInfo();
    inf.v = res.tsize();
    res.setInfo(inf);
    res.setPreThresholdPositions(sparsemask.getPreThresholdPositions());
    res.unthresholdSeries();
    res.writeAsFloat(LogSingleton::getInstance().appendDir("res4d"));
    res.CleanUp();
    
    copedata.CleanUp();
    save_volume(mask, LogSingleton::getInstance().appendDir("mask"),volinfo);

    const VolumeInfo& info = sparsemask.getInfo();
    
    for(int t = 0; t < design.getnumtcontrasts(); t++)
      {
	ts[t].setInfo(info);
	ts[t].setPreThresholdPositions(sparsemask.getPreThresholdPositions());
	ts[t].unthreshold();
	ts[t].writeAsFloat(LogSingleton::getInstance().appendDir("tstat"+num2str(t+1)));
	ts[t].CleanUp();

	tdofs[t].setInfo(info);
	tdofs[t].setPreThresholdPositions(sparsemask.getPreThresholdPositions());
	tdofs[t].unthreshold();
	tdofs[t].writeAsFloat(LogSingleton::getInstance().appendDir("tdof_t"+num2str(t+1)));
	tdofs[t].CleanUp();

	zts[t].setInfo(info);
	zts[t].setPreThresholdPositions(sparsemask.getPreThresholdPositions());
	zts[t].unthreshold();
	zts[t].writeAsFloat(LogSingleton::getInstance().appendDir("zstat"+num2str(t+1)));
	zts[t].CleanUp();

	tcopes[t].setInfo(info);
	tcopes[t].setPreThresholdPositions(sparsemask.getPreThresholdPositions());
	tcopes[t].unthreshold();
	tcopes[t].writeAsFloat(LogSingleton::getInstance().appendDir("cope"+num2str(t+1)));
	tcopes[t].CleanUp();

	tvarcopes[t].setInfo(info);
	tvarcopes[t].setPreThresholdPositions(sparsemask.getPreThresholdPositions());
	tvarcopes[t].unthreshold();
	tvarcopes[t].writeAsFloat(LogSingleton::getInstance().appendDir("varcope"+num2str(t+1)));
	tvarcopes[t].CleanUp();

      }
// 	save_volume(ts[t], LogSingleton::getInstance().appendDir("tstat"+num2str(t+1)),volinfo);
// 	save_volume(tdofs[t], LogSingleton::getInstance().appendDir("tdof_t"+num2str(t+1)),volinfo);
// 	save_volume(zts[t], LogSingleton::getInstance().appendDir("zstat"+num2str(t+1)),volinfo);
// 	save_volume(tcopes[t], LogSingleton::getInstance().appendDir("cope"+num2str(t+1)),volinfo);
// 	save_volume(tvarcopes[t], LogSingleton::getInstance().appendDir("varcope"+num2str(t+1)),volinfo);
//       }

    for(int f = 0; f < design.getnumfcontrasts(); f++)
      {
	fs[f].setInfo(info);
	fs[f].setPreThresholdPositions(sparsemask.getPreThresholdPositions());
	fs[f].unthreshold();
	fs[f].writeAsFloat(LogSingleton::getInstance().appendDir("fstat"+num2str(f+1)));
	fs[f].CleanUp();
	
	fdof1s[f].setInfo(info);
	fdof1s[f].setPreThresholdPositions(sparsemask.getPreThresholdPositions());
	fdof1s[f].unthreshold();
	fdof1s[f].writeAsFloat(LogSingleton::getInstance().appendDir("fdof1_f"+num2str(f+1)));
	fdof1s[f].CleanUp();

	fdof2s[f].setInfo(info);
	fdof2s[f].setPreThresholdPositions(sparsemask.getPreThresholdPositions());
	fdof2s[f].unthreshold();
	fdof2s[f].writeAsFloat(LogSingleton::getInstance().appendDir("fdof2_f"+num2str(f+1)));
	fdof2s[f].CleanUp();

	zfs[f].setInfo(info);
	zfs[f].setPreThresholdPositions(sparsemask.getPreThresholdPositions());
	zfs[f].unthreshold();
	zfs[f].writeAsFloat(LogSingleton::getInstance().appendDir("zfstat"+num2str(f+1)));
	zfs[f].CleanUp();
      }

//     for(int f = 0; f < design.getnumfcontrasts(); f++)
//       {
// 	save_volume(fs[f], LogSingleton::getInstance().appendDir("fstat"+num2str(f+1)),volinfo);
// 	save_volume(fdof1s[f], LogSingleton::getInstance().appendDir("fdof1_f"+num2str(f+1)),volinfo);
// 	save_volume(fdof2s[f], LogSingleton::getInstance().appendDir("fdof2_f"+num2str(f+1)),volinfo);
// 	save_volume(zfs[f], LogSingleton::getInstance().appendDir("zfstat"+num2str(f+1)),volinfo);
//       }

    for(int e = 0; e < nevs; e++)
      {
	pes[e].setInfo(info);
	pes[e].setPreThresholdPositions(sparsemask.getPreThresholdPositions());
	pes[e].unthreshold();
	pes[e].writeAsFloat(LogSingleton::getInstance().appendDir("pe"+num2str(e+1)));
	pes[e].CleanUp();
      }

//     for(int e = 0; e < nevs; e++)
//       {
// 	save_volume(pes[e], LogSingleton::getInstance().appendDir("pe")+num2str(e+1),volinfo);
//       }


    for(int g = 0; g < ngs; g++)
      {
	beta_mean[g].setInfo(info);
	beta_mean[g].setPreThresholdPositions(sparsemask.getPreThresholdPositions());
	beta_mean[g].unthreshold();
	beta_mean[g].writeAsFloat(LogSingleton::getInstance().appendDir("mean_random_effects_var"+num2str(g+1)));
	beta_mean[g].CleanUp();
      }

//     save_volume4D(beta_mean, LogSingleton::getInstance().appendDir("mean_random_effects_var"),volinfo);

    
  }

  void Gsmanager::ols()
  {
    Tracer_Plus trace("Gsmanager::ols");

    const Matrix& dm = design.getdm();

    Matrix pinvdm = (dm.t()*dm).i();

    Matrix pinvdmdm = pinvdm*dm.t();

    if(nevs >= ntpts)
      {
	throw Exception("Singular matrix. ");
      }
 
    int vox = 0;
    for(int x = 0; x < xsize; x++)
      for(int y = 0; y < ysize; y++)
	for(int z = 0; z < zsize; z++)
	  {
	    if(mask(x,y,z))
	      {
		vox++;
		ColumnVector Y = copedata.getSeries(vox);

		ColumnVector petmp = pinvdmdm*Y;

		for(int e = 0; e < nevs; e++)
		  {
		    pes[e](vox) = petmp(e+1);
		  }

		ColumnVector r = Y-dm*petmp;
		float r2 = (r.t()*r).AsScalar();
		
		Matrix covariance = (pinvdm*r2/(ntpts-nevs));
		ColumnVector covariance2;
		reshape(covariance2, covariance, nevs*nevs, 1);

		cov_pes.setSeries(covariance2,vox);

		float betamean = 1.0/(Max(0.01,(r2/(ntpts-1)-mean(varcopedata.getSeries(vox)).AsScalar())));
		float betavar = abs(betamean)*2;
		
		for(int g = 0; g < ngs; g++)
		  {
		    beta_b[g](vox) = Sqr(betamean)/betavar;
		    beta_c[g](vox) = betamean/betavar;
		  }

		if(GsOptions::getInstance().zthreshold.value() > 0)
		  {
		    ols_contrasts(petmp,covariance,vox);
		    zmask(x,y,z) = pass_through_to_mcmc(GsOptions::getInstance().zthreshold.value(),vox);
		  }

	      }
	  }    

    nmaskvoxels = int(zmask.sum());
    
  }
  
  void Gsmanager::ols_contrasts(const ColumnVector& mn, const Matrix& covariance, int vox)
  {
    Tracer_Plus trace("Gsmanager::ols_contrasts");    
    
    for(int t = 0; t < design.getnumtcontrasts(); t++)
      {
	RowVector tcon = design.gettcontrast(t+1);
	t_ols_contrast(mn,covariance,tcon,tcopes[t](vox), tvarcopes[t](vox), ts[t](vox), tdofs[t](vox), zts[t](vox));
      }

    for(int f = 0; f < design.getnumfcontrasts(); f++)
      {
	f_ols_contrast(mn,covariance,design.getfcontrast(f+1),fs[f](vox), fdof1s[f](vox), fdof2s[f](vox), zfs[f](vox));
      }
    
  }

  bool Gsmanager::pass_through_to_mcmc(float zthresh, int vox)
  {
    Tracer_Plus trace("Gsmanager::pass_through_to_mcmc");   
    
    bool largez = false;
    
    for(int t = 0; !largez && t < design.getnumtcontrasts(); t++)
      {	
	largez = (zts[t](vox) > zthresh);
      }
    for(int f = 0; !largez && f < design.getnumfcontrasts(); f++)
      {
	largez = (zfs[f](vox) > zthresh);
      }

    return largez;
  }

  void Gsmanager::t_ols_contrast(const ColumnVector& mn, const Matrix& covariance, const RowVector& tcontrast, double& cope, double& varcope, double& t, double& dof, double& z)
  {
    Tracer_Plus trace("Gsmanager::t_ols_contrast");

    varcope = (tcontrast*covariance*tcontrast.t()).AsScalar();
    cope = (tcontrast*mn).AsScalar();

    dof = float(ntpts - nevs);
    t = cope/sqrt(varcope);

    z = T2z::getInstance().convert(t,int(dof));
  }

  void Gsmanager::f_ols_contrast(const ColumnVector& mn, const Matrix& covariance, const Matrix& fcontrast, double& f, double& dof1, double& dof2, double& z)
  {
    Tracer_Plus trace("Gsmanager::f_ols_contrast");

    dof1 = float(fcontrast.Nrows());
    dof2 = float(ntpts - nevs);
    
    f = (mn.t()*fcontrast.t()*(fcontrast*covariance*fcontrast.t()).i()*fcontrast*mn/dof1).AsScalar();

    z = F2z::getInstance().convert(f,int(dof1),int(dof2));
  }
 
  void Gsmanager::mcmc_contrasts(const Matrix& gamsamples, int vox)
  {
    Tracer_Plus trace("Gsmanager::mcmc_contrasts");   
    
    for(int t = 0; t < design.getnumtcontrasts(); t++)
      {
	RowVector tcon = design.gettcontrast(t+1);
	t_mcmc_contrast(gamsamples, tcon, tcopes[t](vox), tvarcopes[t](vox), ts[t](vox), tdofs[t](vox), zts[t](vox));		    
      }

    for(int f = 0; f < design.getnumfcontrasts(); f++)
      {
	f_mcmc_contrast(gamsamples, design.getfcontrast(f+1), fs[f](vox), fdof1s[f](vox), fdof2s[f](vox), zfs[f](vox));		    
      }
       
  }

  void Gsmanager::t_mcmc_contrast(const Matrix& gamsamples, const RowVector& tcontrast, double& cope, double& varcope, double& t, double& dof, double& z)
  {
    Tracer_Plus trace("Gsmanager::t_mcmc_contrast");
 
    //gamsamples(nevs, nsamples);
  
    ColumnVector m;
    SymmetricMatrix covar;

    Matrix tcsamples = tcontrast*gamsamples;

    float tmpdof;
    multitfit(tcsamples, m, covar, tmpdof);
    dof = double(tmpdof);
 
    varcope = covar(1,1);
    cope = m(1);

    t = cope/sqrt(varcope);
    z = T2z::getInstance().convert(t,int(dof));
  }

  void Gsmanager::f_mcmc_contrast(const Matrix& gamsamples, const Matrix& fcontrast, double& f, double& dof1, double& dof2, double& z)
  {
    Tracer_Plus trace("Gsmanager::f_mcmc_contrast");
    
    //gamsamples(nevs, nsamples);
    
    ColumnVector m;
    SymmetricMatrix covar;
    
    float tmpdof2;
    multitfit(gamsamples, m, covar, tmpdof2);
    dof2 = double(tmpdof2);

    dof1 = float(fcontrast.Nrows());

    f = (m.t()*fcontrast.t()*(fcontrast*covar*fcontrast.t()).i()*fcontrast*m/dof1).AsScalar();

    z = F2z::getInstance().convert(f,int(dof1),int(dof2));
  }

}
